@@ -1,14 +1,15 @@
 // [[Rcpp::interfaces(r, cpp)]]
 //'
 
-#include <boost/math/constants/constants.hpp> // for pi
 #include <fftw3.h>
+#define STRICT_R_HEADERS
 #include <Rcpp.h>
 using namespace Rcpp;
 
 // used to make fft computations more readable
 #define REAL 0
 #define IMAG 1
+//#define PI2 3.14159265358979323846264338327950288419716939937510582097494459230781640628620899862803482534211706798214808651e+00
 
 /******************************************************************/
 /**   Helper Functions                                           **/
@@ -166,19 +167,20 @@ NumericVector dpb_conv(const IntegerVector obs, const NumericVector probs){
   // number of input probabilities
   const int size = probs.length();
   
-  NumericVector temp(size + 1);
+  // results vector
   NumericVector results(size + 1);
   results[0] = 1 - probs[0];
   results[1] = probs[0];
   
   for(int i = 1; i < size; i++){
     checkUserInterrupt();
-    for(int j = 0; j <= i; j++){
-      temp[j] = results[j];
-      if(temp[j]) results[j] *= 1 - probs[i];
-    }
-    for(int j = 0; j <= i; j++){
-      if(temp[j] && probs[i]) results[j + 1] += temp[j] * probs[i];
+    if(probs[i]){
+      for(int j = i; j >= 0; j--){
+        if(results[j]){
+          results[j + 1] += results[j] * probs[i];
+          results[j] *= 1 - probs[i];
+        }
+      }
     }
   }
   // make sure that probability masses sum up to 1
@@ -219,7 +221,7 @@ NumericVector fft_probs(const NumericVector probsA, const NumericVector probsB){
   const int sizeResult = sizeA + sizeB - 1;
   
   // results vector
-  NumericVector result(sizeResult);
+  double *result_vec = new double[sizeResult];
   
   // allocate memory for FFTs of the probs and the convolution result
   fftw_complex *probsA_fft, *probsB_fft, *result_fft;
@@ -227,7 +229,7 @@ NumericVector fft_probs(const NumericVector probsA, const NumericVector probsB){
   // 0-padding of probsA vector and perform FFT of it
   NumericVector padded_probsA(sizeResult);
   padded_probsA[Range(0, sizeA - 1)] = probsA;
-  probsA_fft = (fftw_complex *)fftw_malloc(sizeof(fftw_complex)* sizeResult);
+  probsA_fft = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * sizeResult);
   fftw_plan planA = fftw_plan_dft_r2c_1d(sizeResult, padded_probsA.begin(), probsA_fft, FFTW_ESTIMATE);
   fftw_execute(planA);
   fftw_destroy_plan(planA);
@@ -235,20 +237,21 @@ NumericVector fft_probs(const NumericVector probsA, const NumericVector probsB){
   // 0-padding of probsB vector and perform FFT of it
   NumericVector padded_probsB(sizeResult);
   padded_probsB[Range(0, sizeB - 1)] = probsB;
-  probsB_fft = (fftw_complex *)fftw_malloc(sizeof(fftw_complex)* sizeResult);
+  probsB_fft = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * sizeResult);
   fftw_plan planB = fftw_plan_dft_r2c_1d(sizeResult, padded_probsB.begin(), probsB_fft, FFTW_ESTIMATE);
   fftw_execute(planB);
   fftw_destroy_plan(planB);
   
   // convolute by complex multiplication of the transformed input probs
-  result_fft = (fftw_complex *)fftw_malloc(sizeof(fftw_complex)*sizeResult);
+  result_fft = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * sizeResult);
   for(int i = 0; i < sizeResult; i++){
     result_fft[i][REAL] = (probsA_fft[i][REAL]*probsB_fft[i][REAL] - probsA_fft[i][IMAG]*probsB_fft[i][IMAG])/sizeResult;
     result_fft[i][IMAG] = (probsA_fft[i][REAL]*probsB_fft[i][IMAG] + probsA_fft[i][IMAG]*probsB_fft[i][REAL])/sizeResult;
   }
   
   // inverse tranformation of the above multiplications
-  fftw_plan planResult = fftw_plan_dft_c2r_1d(sizeResult, result_fft, result.begin(), FFTW_ESTIMATE);
+  //fftw_plan planResult = fftw_plan_dft_c2r_1d(sizeResult, result_fft, result.begin(), FFTW_ESTIMATE);
+  fftw_plan planResult = fftw_plan_dft_c2r_1d(sizeResult, result_fft, result_vec, FFTW_ESTIMATE);
   fftw_execute(planResult);
   fftw_destroy_plan(planResult);
   
@@ -258,6 +261,9 @@ NumericVector fft_probs(const NumericVector probsA, const NumericVector probsB){
   fftw_free(result_fft);
   
   // return final results
+  NumericVector result(sizeResult);
+  for(int i = 0; i < sizeResult; i++) result[i] = result_vec[i];
+  delete[] result_vec;
   return result;
 }
 
@@ -266,9 +272,9 @@ NumericVector dpb_dc(const IntegerVector obs, const NumericVector probs){//, con
   // number of probabilities of success
   const int size = probs.length();
   
-  // automatically determine number of splits, if size is above 600
-  //int num_splits = splits < 0 ? std::max<int>(0, (int)std::ceil(std::log(size / 950) / std::log(2.0))) : splits;
-  int num_splits = size > 950 ? (int)std::ceil(std::log(size / 950) / std::log(2.0)) : 0;
+  // automatically determine number of splits, if size is above 1950
+  //int num_splits = splits < 0 ? std::max<int>(0, (int)std::ceil(std::log(size / 1950) / std::log(2.0))) : splits;
+  int num_splits = size > 1950 ? (int)std::ceil(std::log(size / 1950) / std::log(2.0)) : 0;
   // direct convolution is sufficient in case of 0 splits
   if(num_splits == 0) return dpb_conv(obs, probs);
   // number of groups
@@ -391,32 +397,11 @@ NumericVector dpb_dftcf(const IntegerVector obs, const NumericVector probs){
   input_fft[0][IMAG] = 0.0;
   
   // initialize complex numbers for "C" and "C to the power of i"
-  const std::complex<double> C = exp(std::complex<double>(0.0, 2.0) * boost::math::double_constants::pi / ((double)sizeOut));
+  const std::complex<double> C = exp(std::complex<double>(0.0, 2.0) * M_PI / ((double)sizeOut));
   std::complex<double> C_power = 1.0;
   
-  const int mid = sizeIn / 2 + 1;
-  /*double omega = 2 * boost::math::double_constants::pi / ((double)sizeOut);
-  double d, x, c, s;
-  std::complex<double> z;
-  for(int l = 1; l <= mid; l++){
-    d = 0.0;
-    x = 0.0;
-    for(int j = 0; j < sizeIn; j++){
-      z.real(probs[j] * (std::cos(omega * l) - 1) + 1);
-      z.imag(probs[j] * std::sin(omega * l));
-      d += std::log(std::abs(z));
-      x += std::arg(z);
-    }
-    d = std::exp(d);
-    c = d * std::cos(x);
-    s = d * std::sin(x);
-    input_fft[l][REAL] = c;
-    input_fft[l][IMAG] = s;
-    input_fft[sizeOut - l][REAL] = c;
-    input_fft[sizeOut - l][IMAG] = -s;
-  }*/
-  
   // compute closed-form expression of Hernandez and Williams
+  const int mid = sizeIn / 2 + 1;
   for(int i = 1; i <= mid; i++){
     checkUserInterrupt();
     
@@ -624,13 +609,24 @@ NumericVector ppb_gmba(const IntegerVector obs, const NumericVector probs, const
 // Poisson Approximation
 // [[Rcpp::export]]
 NumericVector dpb_pa(const IntegerVector obs, const NumericVector probs){
+  // number of probabilities of success
+  const int size = probs.length();
   // sum of probability is the expectation of the Poisson approximation
   const double lambda = sum(probs);
   
-  // compute probability masses and return
-  if(obs.length() == 0)
-    return dpois(IntegerVector(Range(0, probs.length())), lambda);
-  else return dpois(obs, lambda);
+  // compute probability masses
+  NumericVector results;
+  if(obs.length() == 0){
+    results = dpois(IntegerVector(Range(0, size)), lambda);
+    results[size] += R::ppois(size, lambda, false, false);
+  }else{
+    results = dpois(obs, lambda);
+    for(int i = 0; i < obs.length(); i++)
+      if(obs[i] == size) results[i] += R::ppois(size, lambda, false, false);
+  }
+  
+  // return final results
+  return results;
 }
 
 // [[Rcpp::export]]
@@ -639,11 +635,7 @@ NumericVector ppb_pa(const IntegerVector obs, const NumericVector probs, bool lo
   const double lambda = sum(probs);
   
   // compute cumulative probabilities
-  IntegerVector observed;
-  
-  if(obs.length() == 0)
-    observed = IntegerVector(Range(0, probs.length()));
-  else observed = obs;
+  const IntegerVector observed = obs.length() ? obs : IntegerVector(Range(0, probs.length()));
   
   NumericVector results = ppois(observed, lambda, lower_tail);
   
@@ -687,7 +679,7 @@ NumericVector ppb_na(const IntegerVector obs, const NumericVector probs, const b
   
   // make sure largest possible value has cumulative probability of 1 (lower tail) or 0 (upper tail)
   if(obs.length()){
-    if(max_q == size) results[obs == max_q] = (double)lower_tail;
+    if(max_q >= size) results[obs >= max_q] = (double)lower_tail;
   }else results[size] = (double)lower_tail;
   
   // return final results
@@ -763,43 +755,69 @@ NumericVector dgpb_conv(const IntegerVector obs, const NumericVector probs, cons
   // number of probabilities of success
   const int sizeIn = probs.length();
   // determine pairwise minimum and maximum
-  const IntegerVector u = pmax(val_p, val_q);
   const IntegerVector v = pmin(val_p, val_q);
+  const IntegerVector u = pmax(val_p, val_q);
   // compute differences
-  const IntegerVector d = u - v;
-  // output size
+  IntegerVector d = u - v;
+  // final output size
   const int sizeOut = sum(d) + 1;
+  // greatest common divisor of the differences
+  const int gcd = vectorGCD(d[d > 0]);
+  // rescale differences according to gcd
+  if(gcd > 1) d = d / gcd;
+  // theoretical rescaled maximum
+  const int max_rescaled = sum(d);
+  // output size
+  const int sizeOut_rescaled = max_rescaled + 1;
+  
+  // if val_p[i] was not the larger one, the respective probs[i] has to be 'flipped'
+  // furthermore: if difference is 0 (i.e. u[i] equals v[i]), a non-zero outcome is impossible
+  NumericVector probs_flipped(sizeIn);
+  for(int i = 0; i < sizeIn; i++){
+    if(!d[i]) probs_flipped[i] = 0.0;
+    else if(val_p[i] < u[i]) probs_flipped[i] = 1 - probs[i];
+      else probs_flipped[i] = probs[i];
+  }
   
   // results vectors
   NumericVector results(sizeOut);
-  // initialize results (first convolution step)
-  results[0] = 1.0;
-  // position of last computed probability
-  int end = 0;
+  NumericVector results_rescaled;
   
-  // perform convolution
-  NumericVector temp(sizeOut);
-  double prob;
-  for(int i = 0; i < sizeIn; i++){
-    checkUserInterrupt();
-    if(val_p[i] < u[i]) prob = 1 - probs[i]; else prob = probs[i];
-    
-    for(int j = 0; j <= end; j++){
-      temp[j] = results[j];
-      if(temp[j]) results[j] = temp[j] * (1 - prob);
+  // if maximum rescaled difference equals 1, we have an ordinary poisson binomial distribution
+  if(max(d) == 1){
+    // compute ordinary distribution
+    results_rescaled = dpb_conv(IntegerVector(), probs_flipped[d > 0]);
+  }else{
+    results_rescaled = NumericVector(sizeOut_rescaled);
+    // initialize result of first convolution step
+    results_rescaled[0] = 1.0;
+    // ending position of last computed iteration
+    int end = 0;
+    // perform convolution
+    for(int i = 0; i < sizeIn; i++){
+      checkUserInterrupt();
+      if(probs_flipped[i]){
+        for(int j = end; j >= 0; j--){
+          if(d[i] && results_rescaled[j] && probs_flipped[i]){
+            results_rescaled[j + d[i]] += results_rescaled[j] * probs_flipped[i];
+            results_rescaled[j] *= 1 - probs_flipped[i];
+          }
+        }
+        
+        // update ending position
+        end += d[i];
+      }
     }
-    for(int j = 0; j <= end; j++){
-      if(temp[j] && prob) results[j + d[i]] += temp[j] * prob;
-    }
-    
-    end += d[i];
   }
-  
   // "correct" numerically false (and thus useless) results
-  results[results > 1] = 1;
+  results_rescaled[results_rescaled > 1] = 1;
   
   // make sure that probability masses sum up to 1
-  norm_dpb(results);
+  norm_dpb(results_rescaled);
+  
+  // map results to generalized distribution (scale-back)
+  for(int i = 0; i < sizeOut_rescaled; i++)
+    results[i * gcd] = results_rescaled[i];
   
   // return final results
   if(obs.length()) return results[obs - sum(v)]; else return results;
@@ -837,134 +855,152 @@ NumericVector dgpb_dc(const IntegerVector obs, const NumericVector probs, const
   // determine pairwise minimum and maximum
   IntegerVector v = pmin(val_p, val_q);
   IntegerVector u = pmax(val_p, val_q);
-  // compute differences
-  const IntegerVector d = u - v;
   // theoretical minimum
   const int min_v = sum(v);
+  // compute differences
+  IntegerVector d = u - v;
+  // final output size
+  const int sizeOut = sum(d) + 1;
+  // greatest common divisor of the differences
+  const int gcd = vectorGCD(d[d > 0]);
+  // rescale differences according to gcd
+  if(gcd > 1) d = d / gcd;
+  // theoretical rescaled maximum
+  const int max_rescaled = sum(d);
   // output size
-  const double sizeOut = sum(d);
+  const int sizeOut_rescaled = max_rescaled + 1;
   
-  // number of tree splits
-  //int num_splits = splits < 0 ? std::max<int>(0, (int)round(std::log(sizeOut)/std::log(2.0)/6 + 5 * std::log(sizeIn)/std::log(2.0)/6 - 8.75)) : splits;
-  int num_splits = sizeIn > 0 ? std::max<int>(0, (int)round(0.79 * std::log(sizeIn)/std::log(2.0) + 0.035 * std::log(sizeOut)/std::log(2.0) - 6.85)) : 0;
-  // direct convolution is sufficient in case of 0 splits
-  if(num_splits == 0) return dgpb_conv(obs, probs, val_p, val_q);
-  // number of groups
-  int num_groups = std::pow(2, num_splits);
-  // fraction of total size per group
-  double frac = (double)sizeOut/num_groups;
-  // reduce number of splits and groups if inner-group sizes are too large
-  while(num_splits > 0 && (num_groups > sizeIn || frac < max(d))){
-    num_splits -= 1;
-    num_groups /= 2;
-    frac *= 2;
-  }
-  // direct convolution is sufficient, if no splits are necessary
-  if(num_splits == 0) return dgpb_conv(obs, probs, val_p, val_q);
-  
-  // compute group sizes with minimum size disparity
-  IntegerVector group_sizes(num_groups);
-  IntegerVector group_indices(sizeIn, -1);
-  // reorder 'prob' according to u,v
-  NumericVector probs_ordered(sizeIn);
+  // if val_p[i] was not the larger one, the respective probs[i] has to be 'flipped'
+  // furthermore: if difference is 0 (i.e. u[i] equals v[i]), a non-zero outcome is impossible
+  NumericVector probs_flipped(sizeIn);
   for(int i = 0; i < sizeIn; i++){
-    if(val_p[i] < u[i]) probs_ordered[i] = 1 - probs[i];
-    else probs_ordered[i] = probs[i];
+    if(!d[i]) probs_flipped[i] = 0.0;
+    else if(val_p[i] < u[i]) probs_flipped[i] = 1 - probs[i];
+      else probs_flipped[i] = probs[i];
   }
   
-  // assign each probability and outcome to a group
-  IntegerVector ord = order(NumericVector(d), true);
-  IntegerVector d_ordered = d[ord];
-  probs_ordered = probs_ordered[ord];
-  NumericVector remainder(num_groups, frac);
-  int g = 0;
-  int inc = 1;
-  for(int i = 0; i < sizeIn; i++){
-    checkUserInterrupt();
-    if(g == num_groups || g == -1){
-      inc *= -1;
+  // results vectors
+  NumericVector results(sizeOut);
+  NumericVector results_rescaled;
+  
+  // if max_rescaled equals input size, we have an ordinary poisson binomial distribution
+  if(max(d) == 1){
+    // compute ordinary distribution
+    results_rescaled = dpb_dc(IntegerVector(), probs_flipped);
+  }else{
+    // number of tree splits
+    //int num_splits = splits < 0 ? std::max<int>(0, (int)std::ceil(std::log(sizeIn / 860) / std::log(2.0))) : splits;
+    int num_splits = sizeIn > 860 ? std::max<int>(0, (int)std::ceil(std::log(sizeIn / 860) / std::log(2.0))) : 0;
+    // direct convolution is sufficient in case of 0 splits
+    if(num_splits == 0) return dgpb_conv(obs, probs_flipped, u, v);
+    // number of groups
+    int num_groups = std::pow(2, num_splits);
+    // fraction of total size per group
+    double frac = (double)(sizeOut_rescaled - 1)/num_groups;
+    // reduce number of splits and groups if inner-group sizes are too large
+    while(num_splits > 0 && (num_groups > sizeIn || frac < max(d))){
+      num_splits -= 1;
+      num_groups /= 2;
+      frac *= 2;
+    }
+    // direct convolution is sufficient, if no splits are necessary
+    if(num_splits == 0) return dgpb_conv(obs, probs_flipped, u, v);
+    
+    // compute group sizes with minimum size disparity
+    IntegerVector group_sizes(num_groups);
+    IntegerVector group_indices(sizeIn, -1);
+    
+    // assign each probability and outcome to a group
+    IntegerVector ord = order(NumericVector(d), true);
+    IntegerVector d_ordered = d[ord];
+    probs_flipped = probs_flipped[ord];
+    NumericVector remainder(num_groups, frac);
+    int g = 0;
+    int inc = 1;
+    for(int i = 0; i < sizeIn; i++){
+      checkUserInterrupt();
+      if(g == num_groups || g == -1){
+        inc *= -1;
+        g += inc;
+      }
+      if(d_ordered[i] > remainder[g]){
+        g = 0;
+        for(int j = 1; j < num_groups; j++){
+          if(remainder[j] > remainder[g]) g = j;
+        }
+      }
+      group_sizes[g] += d_ordered[i];
+      remainder[g] -= d_ordered[i];
+      group_indices[i] = g;
       g += inc;
     }
-    if(d_ordered[i] > remainder[g]){
-      g = 0;
-      for(int j = 1; j < num_groups; j++){
-        if(remainder[j] > remainder[g]) g = j;
-      }
+    
+    // compute first and last indices of the groups
+    IntegerVector group_starts(num_groups);
+    IntegerVector group_ends(num_groups);
+    group_starts[0] = 0;
+    group_ends[0] = group_sizes[0] - 1;
+    for(int i = 1; i < num_groups; i++){
+      group_starts[i] = group_starts[i - 1] + group_sizes[i - 1];
+      group_ends[i] = group_ends[i - 1] + group_sizes[i];
     }
-    group_sizes[g] += d_ordered[i];
-    remainder[g] -= d_ordered[i];
-    group_indices[i] = g;
-    g += inc;
-  }
-  
-  // compute first and last indices of the groups
-  IntegerVector group_starts(num_groups);
-  IntegerVector group_ends(num_groups);
-  group_starts[0] = 0;
-  group_ends[0] = group_sizes[0] - 1;
-  for(int i = 1; i < num_groups; i++){
-    group_starts[i] = group_starts[i - 1] + group_sizes[i - 1];
-    group_ends[i] = group_ends[i - 1] + group_sizes[i];
-  }
-  
-  // results vector; direct convolution will increase size of each group by 1
-  NumericVector results(sizeOut + num_groups);
-  
-  // compute direct convolutions for each group
-  int start = 0, end = 0;
-  for(int i = 0; i < num_groups; i++){
-    checkUserInterrupt();
-    // compute new starting and ending indices, because groups grow by 1
-    start = group_starts[i] + i;
-    end = group_ends[i] + i + 1;
     
-    // target range
-    Range target(start, end);
-    ////u = val_p[Range(probs_start[i], probs_end[i])];
-    ////v = val_q[Range(probs_start[i], probs_end[i])];
-    u = d_ordered[group_indices == i];
-    v = IntegerVector(u.length());
+    // results vector; direct convolution will increase size of each group by 1
+    results_rescaled = NumericVector(sizeOut_rescaled - 1 + num_groups);
     
-    // direct convolution
-    ////results[target] = dgpb_conv(IntegerVector(), probs[Range(probs_start[i], probs_end[i])], u, v);
-    results[target] = dgpb_conv(IntegerVector(), probs_ordered[group_indices == i], u, v);
-    
-    // update starting and ending positions
-    group_starts[i] = start;
-    group_ends[i] = end;
-  }
-  
-  int num_groups_reduced = num_groups / 2;
-  while(num_splits > 0){
-    for(int i = 0; i < num_groups_reduced; i++){
+    // compute direct convolutions for each group
+    int start = 0, end = 0;
+    for(int i = 0; i < num_groups; i++){
       checkUserInterrupt();
-      // compute new starting and ending indices, because group sizes are
-      // reduced by 1, due to FFT convolution
-      start = group_starts[2*i] - i;
-      end = group_ends[2*i + 1] - i - 1;
+      // compute new starting and ending indices, because groups grow by 1
+      start = group_starts[i] + i;
+      end = group_ends[i] + i + 1;
       
       // target range
       Range target(start, end);
-      // FFT convolution
-      results[target] = fft_probs(results[Range(group_starts[2*i], group_ends[2*i])], results[Range(group_starts[2*i + 1], group_ends[2*i + 1])]);
+      u = d_ordered[group_indices == i];
+      v = IntegerVector(u.length());
       
-      // update starting and ending indices
+      // direct convolution
+      results_rescaled[target] = dgpb_conv(IntegerVector(), probs_flipped[group_indices == i], u, v);
+      
+      // update starting and ending positions
       group_starts[i] = start;
       group_ends[i] = end;
     }
-    num_groups_reduced /= 2;
-    num_splits -= 1;
+    
+    int num_groups_reduced = num_groups / 2;
+    while(num_splits > 0){
+      for(int i = 0; i < num_groups_reduced; i++){
+        checkUserInterrupt();
+        // compute new starting and ending indices, because group sizes are
+        // reduced by 1, due to FFT convolution
+        start = group_starts[2*i] - i;
+        end = group_ends[2*i + 1] - i - 1;
+        
+        // target range
+        Range target(start, end);
+        // FFT convolution
+        results_rescaled[target] = fft_probs(results_rescaled[Range(group_starts[2*i], group_ends[2*i])], results_rescaled[Range(group_starts[2*i + 1], group_ends[2*i + 1])]);
+        
+        // update starting and ending indices
+        group_starts[i] = start;
+        group_ends[i] = end;
+      }
+      num_groups_reduced /= 2;
+      num_splits -= 1;
+    }
   }
-  
-  // select final results
-  results = results[Range(0, sizeOut)];
-  
   // "correct" numerically false (and thus useless) results
-  results[results < 5.55e-17] = 0;
-  results[results > 1] = 1;
+  results_rescaled[results_rescaled < 5.55e-17] = 0;
+  results_rescaled[results_rescaled > 1] = 1;
   
   // make sure that probability masses sum up to 1
-  norm_dpb(results);
+  norm_dpb(results_rescaled);
+  
+  // map results to generalized distribution (scale-back)
+  for(int i = 0; i < sizeOut_rescaled; i++)
+    results[i * gcd] = results_rescaled[i];
   
   // return final results
   if(obs.length()) return results[obs - min_v]; else return results;
@@ -987,7 +1023,7 @@ NumericVector pgpb_dc(const IntegerVector obs, const NumericVector probs, const
   
   // ensure that sum = 1 (or 0), if last value equals the highest observable value
   if(obs.length()){
-    if(max_q == max_v) results[obs == max_q] = (double)lower_tail;
+    if(max_q == max_v) results[obs == max_v] = (double)lower_tail;
   }else results[max_v - min_v] = (double)lower_tail;
   
   // return final results
@@ -1000,89 +1036,100 @@ NumericVector dgpb_dftcf(const IntegerVector obs, const NumericVector probs, con
   // number of probabilities of success
   const int sizeIn = probs.length();
   // determine pairwise minimum and maximum
-  const IntegerVector u = pmax(val_p, val_q);
   const IntegerVector v = pmin(val_p, val_q);
+  const IntegerVector u = pmax(val_p, val_q);
   // compute differences
-  const IntegerVector d = u - v;
-  // theoretical minimum
-  const int min_v = sum(v);
-  // output size
+  IntegerVector d = u - v;
+  // final output size
   const int sizeOut = sum(d) + 1;
+  // greatest common divisor of the differences
+  const int gcd = vectorGCD(d[d > 0]);
+  // rescale differences according to gcd
+  if(gcd > 1) d = d / gcd;
+  // theoretical rescaled maximum
+  const int max_rescaled = sum(d);
+  // output size
+  const int sizeOut_rescaled = max_rescaled + 1;
   
-  // "initialize" DFT input vector
-  fftw_complex *input_fft;
-  input_fft = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * sizeOut);
-  input_fft[0][REAL] = 1.0;
-  input_fft[0][IMAG] = 0.0;
-  
-  // initialize complex numbers for "C" and "C to the power of i"
-  //std::complex<double> C = 1.0;////
-  //if(min_v) C = exp(std::complex<double>(0.0, -min_v * 2.0) * boost::math::double_constants::pi / ((double)sizeOut));////
-  //std::complex<double> C_power = 1.0;////
-  std::vector< std::complex<double> > Cp(sizeIn, 1.0);
-  //std::vector< std::complex<double> > Cq(sizeIn, 1.0);////
-  std::vector< std::complex<double> > Cp_power(sizeIn, 1.0);
-  //std::vector< std::complex<double> > Cq_power(sizeIn, 1.0);////
+  // if val_p[i] was not the larger one, the respective probs[i] has to be 'flipped'
+  // furthermore: if difference is 0 (i.e. u[i] equals v[i]), a non-zero outcome is impossible
+  NumericVector probs_flipped(sizeIn);
   for(int i = 0; i < sizeIn; i++){
-    if(d[i])
-    //  Cp[i] = exp(std::complex<double>(0.0, val_p[i] * 2.0) * boost::math::double_constants::pi / ((double)sizeOut));////
-      Cp[i] = exp(std::complex<double>(0.0, d[i] * 2.0) * boost::math::double_constants::pi / ((double)sizeOut));
-    //if(val_q[i])////
-    //  Cq[i] = exp(std::complex<double>(0.0, val_q[i] * 2.0) * boost::math::double_constants::pi / ((double)sizeOut));////
+    if(!d[i]) probs_flipped[i] = 0.0;
+    else if(val_p[i] < u[i]) probs_flipped[i] = 1 - probs[i];
+      else probs_flipped[i] = probs[i];
   }
   
-  // compute closed-form expression of Hernandez and Williams
-  double prob;
-  for(int l = 1; l <= sizeOut / 2; l++){
-    checkUserInterrupt();
-    std::complex<double> product = 1.0;
-    for(int k = 0; k < sizeIn; k++){
-      if(val_p[k] < u[k]) prob = 1 - probs[k]; else prob = probs[k];
-      //if(val_p[k]) Cp_power[k] *= Cp[k];////
-      if(d[k]) Cp_power[k] *= Cp[k];////
-      //if(val_q[k]) Cq_power[k] *= Cq[k];////
-      //product *= (probs[k] * Cp_power[k] + (1.0 - probs[k]) * Cq_power[k]);////
-      product *= 1.0 + prob * (Cp_power[k] - 1.0);
+  // results vectors
+  NumericVector results(sizeOut);
+  NumericVector results_rescaled;
+  
+  // if max_rescaled equals input size, we have an ordinary poisson binomial distribution
+  if(max(d) == 1){
+    // compute ordinary distribution
+    results_rescaled = dpb_dftcf(IntegerVector(), probs_flipped[d > 0]);
+  }else{
+    // "initialize" DFT input vector
+    fftw_complex *input_fft;
+    input_fft = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * sizeOut_rescaled);
+    input_fft[0][REAL] = 1.0;
+    input_fft[0][IMAG] = 0.0;
+    
+    // initialize complex numbers for "C" and "C to the power of i"
+    std::vector< std::complex<double> > C(sizeIn, 1.0);
+    std::vector< std::complex<double> > C_power(sizeIn, 1.0);
+    for(int i = 0; i < sizeIn; i++){
+      if(d[i]) C[i] = exp(std::complex<double>(0.0, d[i] * 2.0) * M_PI / ((double)sizeOut_rescaled));
     }
     
-    //if(min_v){////
-    //  C_power *= C;////
-    //  product *= C_power;////
-    //}////
+    // compute closed-form expression of Hernandez and Williams
+    for(int l = 1; l <= sizeOut_rescaled / 2; l++){
+      checkUserInterrupt();
+      std::complex<double> product = 1.0;
+      for(int k = 0; k < sizeIn; k++){
+        if(probs_flipped[k]){
+          if(d[k]) C_power[k] *= C[k];////
+          product *= 1.0 + probs_flipped[k] * (C_power[k] - 1.0);
+        }
+      }
+      
+      input_fft[l][REAL] = product.real();
+      input_fft[l][IMAG] = product.imag();
+      input_fft[sizeOut_rescaled - l][REAL] = product.real();
+      input_fft[sizeOut_rescaled - l][IMAG] = -product.imag();
+    }
     
-    input_fft[l][REAL] = product.real();
-    input_fft[l][IMAG] = product.imag();
-    input_fft[sizeOut - l][REAL] = product.real();
-    input_fft[sizeOut - l][IMAG] = -product.imag();
+    // vector of DFT results
+    fftw_complex *result_fft;
+    result_fft = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * sizeOut_rescaled);
+    
+    // perform DFT
+    fftw_plan planDFT;
+    planDFT = fftw_plan_dft_1d(sizeOut_rescaled, input_fft, result_fft, FFTW_FORWARD, FFTW_ESTIMATE);
+    fftw_execute(planDFT);
+    
+    // gather results
+    results_rescaled = NumericVector(sizeOut_rescaled);
+    for(int i = 0; i < sizeOut_rescaled; i++) results_rescaled[i] = result_fft[i][REAL] / sizeOut_rescaled;
+    
+    // garbage collection
+    fftw_destroy_plan(planDFT);
+    fftw_free(input_fft);
+    fftw_free(result_fft);
   }
-  
-  // vector of DFT results
-  fftw_complex *result_fft;
-  result_fft = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * sizeOut);
-  
-  // perform DFT
-  fftw_plan planDFT;
-  planDFT = fftw_plan_dft_1d(sizeOut, input_fft, result_fft, FFTW_FORWARD, FFTW_ESTIMATE);
-  fftw_execute(planDFT);
-  
-  // gather results
-  NumericVector results(sizeOut);
-  for(int i = 0; i < sizeOut; i++) results[i] = result_fft[i][REAL] / sizeOut;
-  
-  // garbage collection
-  fftw_destroy_plan(planDFT);
-  fftw_free(input_fft);
-  fftw_free(result_fft);
-  
   // "correct" numerically false (and thus useless) results
-  results[results < 2.22e-16] = 0;
-  results[results > 1] = 1;
+  results_rescaled[results_rescaled < 2.22e-16] = 0;
+  results_rescaled[results_rescaled > 1] = 1;
   
   // make sure that probability masses sum up to 1
-  norm_dpb(results);
+  norm_dpb(results_rescaled);
+  
+  // map results to generalized distribution (scale-back)
+  for(int i = 0; i < sizeOut_rescaled; i++)
+    results[i * gcd] = results_rescaled[i];
   
   // return final results
-  if(obs.length()) return results[obs - min_v]; else return results;
+  if(obs.length()) return results[obs - sum(v)]; else return results;
 }
 
 // [[Rcpp::export]]
@@ -1102,7 +1149,7 @@ NumericVector pgpb_dftcf(const IntegerVector obs, const NumericVector probs, con
   
   // ensure that sum = 1 (or 0), if last value equals the highest observable value
   if(obs.length()){
-    if(max_q == max_v) results[obs == max_q] = (double)lower_tail;
+    if(max_q == max_v) results[obs == max_v] = (double)lower_tail;
   }else results[max_v - min_v] = (double)lower_tail;
   
   // return final results
@@ -1111,44 +1158,79 @@ NumericVector pgpb_dftcf(const IntegerVector obs, const NumericVector probs, con
 
 // [[Rcpp::export]]
 NumericVector pgpb_na(const IntegerVector obs, const NumericVector probs, const IntegerVector val_p, const IntegerVector val_q, const bool refined = true, const bool lower_tail = true){
-  // theoretical maximum
-  const int max_v = sum(pmax(val_p, val_q));
-  // maximum observed value
-  const int max_q = obs.length() ? max(obs) : max_v;
-  // mu
-  const double mu = sum(probs * NumericVector(val_p) + (1 - probs) * NumericVector(val_q));
-  // p * q
-  const NumericVector pq = probs * (1 - probs);
-  // sigma
-  const double sigma = std::sqrt(sum(pq * pow(NumericVector(val_p) - NumericVector(val_q), 2.0)));
-  // standardized observations with continuity correction
-  IntegerVector observed;
-  if(obs.length() == 0)
-    observed = IntegerVector(Range(sum(pmin(val_p, val_q)), max_v));
-  else observed = obs;
-  const NumericVector obs_std = (NumericVector(observed) + 0.5 - mu)/sigma;
-  // vector to store results
-  NumericVector results = pnorm(obs_std, 0.0, 1.0, lower_tail);
-  // cumulative probabilities
-  if(refined && sigma){
-    // gamma
-    const double gamma = sum(pq * (1 - 2 * probs) * pow(NumericVector(val_p) - NumericVector(val_q), 3.0))/std::pow(sigma, 3.0);
-    // probabilities
-    if(lower_tail)
-      results += gamma * (1 - pow(obs_std, 2.0)) * dnorm(obs_std) / 6;
-    else
-      results += -gamma * (1 - pow(obs_std, 2.0)) * dnorm(obs_std) / 6;
+  // number of probabilities of success
+  const int sizeIn = probs.length();
+  // determine pairwise minimum and maximum
+  const IntegerVector v = pmin(val_p, val_q);
+  const IntegerVector u = pmax(val_p, val_q);
+  // theoretical (unshifted!) minimum
+  const int min_v = sum(v);
+  // compute differences
+  IntegerVector d = u - v;
+  // maximum observable value
+  const int max_q = sum(d);
+  // greatest common divisor of the differences
+  const int gcd = vectorGCD(d[d > 0]);
+  // rescale differences according to gcd
+  if(gcd > 1) d = d / gcd;
+  // theoretical (shifted!) maximum
+  const int max_rescaled = max_q / gcd;
+  // rescaled observations
+  const IntegerVector obs_range = obs.length() ? (obs - min_v) / gcd : IntegerVector(Range(0, max_rescaled));
+  
+  // if val_p[i] was not the larger one, the respective probs[i] has to be 'flipped'
+  // furthermore: if difference is 0 (i.e. u[i] equals v[i]), a non-zero outcome is impossible
+  NumericVector probs_flipped(sizeIn);
+  for(int i = 0; i < sizeIn; i++){
+    if(!d[i]) probs_flipped[i] = 0.0;
+    else if(val_p[i] < u[i]) probs_flipped[i] = 1 - probs[i];
+    else probs_flipped[i] = probs[i];
   }
-  // make sure that all probabilities do not exceed 1 and are at least 0
-  results[results < 0] = 0;
-  results[results > 1] = 1;
   
-  // make sure largest possible value has cumulative probability of 1
-  if(max_q == max_v)
-    results[observed == max_q] = (double)lower_tail;
-  
-  // return final results
-  return results;
+  // if maximum difference equals 1, we have an ordinary poisson binomial distribution
+  if(max(d) == 1){
+    // compute ordinary distribution
+    return ppb_na(obs_range, probs_flipped[d > 0], refined, lower_tail);
+  }else{
+    // mu
+    const double mu = sum(probs_flipped * NumericVector(d));
+    // p * q
+    const NumericVector pq = probs_flipped * (1 - probs_flipped);
+    // sigma
+    const double sigma = std::sqrt(sum(pq * pow(NumericVector(d), 2.0)));
+    // standardized observations with continuity correction
+    const NumericVector obs_std = (NumericVector(obs_range) + 0.5 - mu)/sigma;
+    // cumulative probabilities
+    NumericVector results_rescaled = pnorm(obs_std, 0.0, 1.0, lower_tail);
+    // refine
+    if(refined && sigma){
+      // gamma
+      const double gamma = sum(pq * (1 - 2 * probs_flipped) * pow(NumericVector(d), 3.0))/std::pow(sigma, 3.0);
+      // probabilities
+      if(lower_tail)
+        results_rescaled += gamma * (1 - pow(obs_std, 2.0)) * dnorm(obs_std) / 6;
+      else
+        results_rescaled += -gamma * (1 - pow(obs_std, 2.0)) * dnorm(obs_std) / 6;
+    }
+    
+    // make sure that all probabilities do not exceed 1 and are at least 0
+    results_rescaled[results_rescaled < 0] = 0;
+    results_rescaled[results_rescaled > 1] = 1;
+    
+    // make sure largest possible value has cumulative probability of 1
+    if(max_q >= max_rescaled) results_rescaled[obs_range >= max_rescaled] = (double)lower_tail;
+    
+    // return final results
+    if(obs.length() || gcd == 1) return results_rescaled;
+    else{
+      // map results to generalized distribution (scale-back)
+      NumericVector results(max_q + 1);
+      for(int i = 0; i <= max_q; i++)
+          results[i] = results_rescaled[i/gcd];
+      
+      return results;
+    }
+  }
 }
 
 // Generalized Normal Approximations (G-NA, G-RNA)
